@@ -1,7 +1,7 @@
 /*
  * dLeyna
  *
- * Copyright (C) 2012-2013 Intel Corporation. All rights reserved.
+ * Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU Lesser General Public License,
@@ -622,7 +622,7 @@ static void prv_add_variant_prop(GVariantBuilder *vb, const gchar *key,
 
 void dls_props_add_child_count(GVariantBuilder *item_vb, gint value)
 {
-	prv_add_int_prop(item_vb, DLS_INTERFACE_PROP_CHILD_COUNT, value);
+	prv_add_uint_prop(item_vb, DLS_INTERFACE_PROP_CHILD_COUNT, value);
 }
 
 static void prv_add_bool_prop(GVariantBuilder *vb, const gchar *key,
@@ -715,7 +715,9 @@ static GVariant *prv_get_artists_prop(GList *list)
 	return g_variant_builder_end(&vb);
 }
 
-void dls_props_add_device(GUPnPDeviceInfo *proxy,
+void dls_props_add_device(GUPnPDeviceInfo *root_proxy,
+			  GUPnPDeviceInfo *proxy,
+			  GUPnPServiceProxy *ems_proxy,
 			  const dls_device_t *device,
 			  GVariantBuilder *vb)
 {
@@ -729,6 +731,10 @@ void dls_props_add_device(GUPnPDeviceInfo *proxy,
 	prv_add_string_prop(vb, DLS_INTERFACE_PROP_UDN,
 			    gupnp_device_info_get_udn(proxy));
 
+	if (proxy != root_proxy)
+		prv_add_string_prop(vb, DLS_INTERFACE_PROP_ROOT_UDN,
+				    gupnp_device_info_get_udn(root_proxy));
+
 	prv_add_string_prop(vb, DLS_INTERFACE_PROP_DEVICE_TYPE,
 			    gupnp_device_info_get_device_type(proxy));
 
@@ -804,7 +810,8 @@ void dls_props_add_device(GUPnPDeviceInfo *proxy,
 				      device->feature_list);
 }
 
-GVariant *dls_props_get_device_prop(GUPnPDeviceInfo *proxy,
+GVariant *dls_props_get_device_prop(GUPnPDeviceInfo *root_proxy,
+				    GUPnPDeviceInfo *proxy,
 				    const dls_device_t *device,
 				    const gchar *prop)
 {
@@ -818,6 +825,9 @@ GVariant *dls_props_get_device_prop(GUPnPDeviceInfo *proxy,
 		str = gupnp_device_info_get_location(proxy);
 	} else if (!strcmp(DLS_INTERFACE_PROP_UDN, prop)) {
 		str = gupnp_device_info_get_udn(proxy);
+	} else if (!strcmp(DLS_INTERFACE_PROP_ROOT_UDN, prop)) {
+		if (proxy != root_proxy)
+			str = gupnp_device_info_get_udn(root_proxy);
 	} else if (!strcmp(DLS_INTERFACE_PROP_DEVICE_TYPE, prop)) {
 		str = gupnp_device_info_get_device_type(proxy);
 	} else if (!strcmp(DLS_INTERFACE_PROP_FRIENDLY_NAME, prop)) {
@@ -1495,6 +1505,23 @@ void dls_props_add_container(GVariantBuilder *item_vb,
 		prv_add_resources(item_vb, GUPNP_DIDL_LITE_OBJECT(object),
 				  filter_mask, FALSE);
 
+
+	if (filter_mask & DLS_UPNP_MASK_PROP_ARTIST) {
+		GUPnPDIDLLiteObject *obj = GUPNP_DIDL_LITE_OBJECT(object);
+		const char *artist = gupnp_didl_lite_object_get_artist(obj);
+		prv_add_string_prop (item_vb,
+				     DLS_INTERFACE_PROP_ARTIST,
+				     artist);
+	}
+
+	if (filter_mask & DLS_UPNP_MASK_PROP_ALBUM_ART_URL) {
+		GUPnPDIDLLiteObject *obj = GUPNP_DIDL_LITE_OBJECT(object);
+		const char *url = gupnp_didl_lite_object_get_album_art(obj);
+		prv_add_string_prop (item_vb,
+				     DLS_INTERFACE_PROP_ALBUM_ART_URL,
+				     url);
+	}
+
 	res = prv_get_matching_resource(GUPNP_DIDL_LITE_OBJECT(object),
 					protocol_info);
 	if (res) {
@@ -2029,6 +2056,20 @@ GVariant *dls_props_get_container_prop(const gchar *prop,
 		DLEYNA_LOG_DEBUG("Prop %s = %u", prop, uint_val);
 
 		retval = g_variant_ref_sink(g_variant_new_uint32(uint_val));
+	} else if (!strcmp(prop, DLS_INTERFACE_PROP_ARTIST)) {
+		const char *strval = gupnp_didl_lite_object_get_artist(
+						GUPNP_DIDL_LITE_OBJECT(object));
+
+		DLEYNA_LOG_DEBUG("Prop %s = %s", prop, strval);
+
+		retval = g_variant_ref_sink(g_variant_new_string(strval));
+	} else if (!strcmp(prop, DLS_INTERFACE_PROP_ALBUM_ART_URL)) {
+		const char *strval = gupnp_didl_lite_object_get_album_art(
+						GUPNP_DIDL_LITE_OBJECT(object));
+
+		DLEYNA_LOG_DEBUG("Prop %s = %s", prop, strval);
+
+		retval = g_variant_ref_sink(g_variant_new_string(strval));
 	} else if (!strcmp(prop, DLS_INTERFACE_PROP_RESOURCES)) {
 		retval = g_variant_ref_sink(
 			prv_compute_resources(object, DLS_UPNP_MASK_ALL_PROPS,
