@@ -1,7 +1,7 @@
 /*
  * dLeyna
  *
- * Copyright (C) 2012-2013 Intel Corporation. All rights reserved.
+ * Copyright (C) 2012-2017 Intel Corporation. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU Lesser General Public License,
@@ -38,6 +38,8 @@
 #include "sort.h"
 #include "upnp.h"
 
+#define DLS_DMS_DEVICE_TYPE "urn:schemas-upnp-org:device:MediaServer:"
+
 struct dls_upnp_t_ {
 	dleyna_connector_id_t connection;
 	const dleyna_connector_dispatch_cb_t *interface_info;
@@ -47,9 +49,9 @@ struct dls_upnp_t_ {
 	dls_upnp_callback_t lost_server;
 	GUPnPContextManager *context_manager;
 	void *user_data;
-	GHashTable *server_udn_map;
-	GHashTable *server_uc_map;
-	guint counter;
+	GHashTable *device_udn_map;
+	GHashTable *sleeping_device_udn_map;
+	GHashTable *device_uc_map;
 };
 
 /* Private structure used in service task */
@@ -84,13 +86,13 @@ static void prv_device_chain_end(gboolean cancelled, gpointer data)
 		goto on_clear;
 
 	DLEYNA_LOG_DEBUG("Notify new server available: %s", device->path);
-	g_hash_table_insert(priv_t->upnp->server_udn_map, g_strdup(priv_t->udn),
+	g_hash_table_insert(priv_t->upnp->device_udn_map, g_strdup(priv_t->udn),
 			    device);
 	priv_t->upnp->found_server(device->path, priv_t->upnp->user_data);
 
 on_clear:
 
-	g_hash_table_remove(priv_t->upnp->server_uc_map, priv_t->udn);
+	g_hash_table_remove(priv_t->upnp->device_uc_map, priv_t->udn);
 
 	if (cancelled)
 		dls_device_delete(device);
@@ -106,7 +108,7 @@ static void prv_device_context_switch_end(gboolean cancelled, gpointer data)
 
 	DLEYNA_LOG_DEBUG("Enter");
 
-	g_hash_table_remove(priv_t->upnp->server_uc_map, priv_t->udn);
+	g_hash_table_remove(priv_t->upnp->device_uc_map, priv_t->udn);
 	prv_device_new_free(priv_t);
 
 	DLEYNA_LOG_DEBUG("Exit");
@@ -146,10 +148,49 @@ static void prv_update_device_context(prv_device_new_ct_t *priv_t,
 	priv_t->queue_id = queue_id;
 	priv_t->device = device;
 
-	g_hash_table_insert(upnp->server_uc_map, g_strdup(udn), priv_t);
+	g_hash_table_insert(upnp->device_uc_map, g_strdup(udn), priv_t);
 }
 
-static void prv_server_available_cb(GUPnPControlPoint *cp,
+static GUPnPDeviceInfo *prv_lookup_dms_child_device(GUPnPDeviceInfo *proxy)
+{
+	GList *child_devices;
+	GList *next;
+	const gchar *device_type;
+	GUPnPDeviceInfo *info = NULL;
+	GUPnPDeviceInfo *child_info = NULL;
+
+	child_devices = gupnp_device_info_list_device_types(proxy);
+
+	next = child_devices;
+	while (next != NULL) {
+		device_type = (gchar *)next->data;
+
+		child_info = gupnp_device_info_get_device(proxy, device_type);
+
+		if (g_str_has_prefix(device_type, DLS_DMS_DEVICE_TYPE)) {
+			break;
+		} else {
+			info = prv_lookup_dms_child_device(child_info);
+
+			g_object_unref(child_info);
+			child_info = NULL;
+
+			if (info != NULL) {
+				child_info = info;
+
+				break;
+			}
+		}
+
+		next = g_list_next(next);
+	}
+
+	g_list_free_full(child_devices, (GDestroyNotify)g_free);
+
+	return child_info;
+}
+
+static void prv_device_available_cb(GUPnPControlPoint *cp,
 				    GUPnPDeviceProxy *proxy,
 				    gpointer user_data)
 {
@@ -161,11 +202,17 @@ static void prv_server_available_cb(GUPnPControlPoint *cp,
 	const dleyna_task_queue_key_t *queue_id;
 	unsigned int i;
 	prv_device_new_ct_t *priv_t;
+	GUPnPDeviceInfo *device_proxy = (GUPnPDeviceInfo *)proxy;
+	GUPnPDeviceInfo *device_info = NULL;
+	const gchar *device_type;
+	gboolean subscribe = FALSE;
+	gpointer key;
+	gpointer val;
 
-	udn = gupnp_device_info_get_udn((GUPnPDeviceInfo *)proxy);
+	udn = gupnp_device_info_get_udn(device_proxy);
 
-	ip_address = gupnp_context_get_host_ip(
-		gupnp_control_point_get_context(cp));
+	ip_address = gssdp_client_get_host_ip(
+		GSSDP_CLIENT(gupnp_control_point_get_context(cp)));
 
 	if (!udn || !ip_address)
 		goto on_error;
@@ -173,10 +220,55 @@ static void prv_server_available_cb(GUPnPControlPoint *cp,
 	DLEYNA_LOG_DEBUG("UDN %s", udn);
 	DLEYNA_LOG_DEBUG("IP Address %s", ip_address);
 
-	device = g_hash_table_lookup(upnp->server_udn_map, udn);
+	device_type = gupnp_device_info_get_device_type(device_proxy);
+
+	if (!g_str_has_prefix(device_type, DLS_DMS_DEVICE_TYPE)) {
+		device_info = prv_lookup_dms_child_device(device_proxy);
+
+		if (device_info == NULL)
+			goto on_error;
+	} else {
+		device_info = device_proxy;
+	}
+
+	device = g_hash_table_lookup(upnp->device_udn_map, udn);
 
 	if (!device) {
-		priv_t = g_hash_table_lookup(upnp->server_uc_map, udn);
+		device = g_hash_table_lookup(upnp->sleeping_device_udn_map,
+					     udn);
+
+		if (device != NULL) {
+			if (g_hash_table_lookup_extended(
+						  upnp->sleeping_device_udn_map,
+						  udn,
+						  &key,
+						  &val)) {
+				g_hash_table_steal(
+						upnp->sleeping_device_udn_map,
+						udn);
+
+				g_free(key);
+			}
+
+			g_hash_table_insert(upnp->device_udn_map, g_strdup(udn),
+					    device);
+
+			if (device->wake_on_timeout_id) {
+				DLEYNA_LOG_DEBUG("Stop WAKE-ON watcher...");
+
+				(void) g_source_remove(
+						 device->wake_on_timeout_id);
+				device->wake_on_timeout_id = 0;
+			}
+			dls_device_delete_context(device->sleeping_context);
+			device->sleeping_context = NULL;
+			device->sleeping = FALSE;
+			subscribe = TRUE;
+		}
+	}
+
+	if (!device) {
+		priv_t = g_hash_table_lookup(upnp->device_uc_map, udn);
 
 		if (priv_t)
 			device = priv_t->device;
@@ -188,15 +280,17 @@ static void prv_server_available_cb(GUPnPControlPoint *cp,
 
 		queue_id = prv_create_device_queue(&priv_t);
 
-		device = dls_device_new(upnp->connection, proxy, ip_address,
+		device = dls_device_new(upnp->connection,
+					proxy,
+					device_info,
+					ip_address,
 					upnp->interface_info,
-					upnp->property_map, upnp->counter,
+					upnp->property_map, 
+					udn,
 					queue_id);
 
 		prv_update_device_context(priv_t, upnp, udn, device, ip_address,
 					  queue_id);
-
-		upnp->counter++;
 	} else {
 		DLEYNA_LOG_DEBUG("Device Found");
 
@@ -208,8 +302,12 @@ static void prv_server_available_cb(GUPnPControlPoint *cp,
 
 		if (i == device->contexts->len) {
 			DLEYNA_LOG_DEBUG("Adding Context");
-			(void) dls_device_append_new_context(device, ip_address,
-							     proxy);
+			(void) dls_device_append_new_context(device,
+							     ip_address,
+							     proxy,
+							     device_info);
+			if (subscribe)
+				dls_device_subscribe_to_service_changes(device);
 		}
 
 		DLEYNA_LOG_DEBUG_NL();
@@ -220,17 +318,17 @@ on_error:
 	return;
 }
 
-static gboolean prv_subscribe_to_contents_change(gpointer user_data)
+static gboolean prv_subscribe_to_service_changes(gpointer user_data)
 {
 	dls_device_t *device = user_data;
 
 	device->timeout_id = 0;
-	dls_device_subscribe_to_contents_change(device);
+	dls_device_subscribe_to_service_changes(device);
 
 	return FALSE;
 }
 
-static void prv_server_unavailable_cb(GUPnPControlPoint *cp,
+static void prv_device_unavailable_cb(GUPnPControlPoint *cp,
 				      GUPnPDeviceProxy *proxy,
 				      gpointer user_data)
 {
@@ -245,6 +343,9 @@ static void prv_server_unavailable_cb(GUPnPControlPoint *cp,
 	gboolean under_construction = FALSE;
 	prv_device_new_ct_t *priv_t;
 	const dleyna_task_queue_key_t *queue_id;
+	dls_device_context_t *lost_context;
+	gpointer key;
+	gpointer val;
 
 	DLEYNA_LOG_DEBUG("Enter");
 
@@ -259,10 +360,10 @@ static void prv_server_unavailable_cb(GUPnPControlPoint *cp,
 	DLEYNA_LOG_DEBUG("UDN %s", udn);
 	DLEYNA_LOG_DEBUG("IP Address %s", ip_address);
 
-	device = g_hash_table_lookup(upnp->server_udn_map, udn);
+	device = g_hash_table_lookup(upnp->device_udn_map, udn);
 
 	if (!device) {
-		priv_t = g_hash_table_lookup(upnp->server_uc_map, udn);
+		priv_t = g_hash_table_lookup(upnp->device_uc_map, udn);
 
 		if (priv_t) {
 			device = priv_t->device;
@@ -284,18 +385,56 @@ static void prv_server_unavailable_cb(GUPnPControlPoint *cp,
 	if (i >= device->contexts->len)
 		goto on_error;
 
-	subscribed = context->subscribed;
+	subscribed = (context->cds.subscribed || context->ems.subscribed);
 	if (under_construction)
 		construction_ctx = !strcmp(context->ip_address,
 					   priv_t->ip_address);
 
-	(void) g_ptr_array_remove_index(device->contexts, i);
+	g_ptr_array_set_free_func(device->contexts, NULL);
+
+	lost_context = g_ptr_array_remove_index(device->contexts, i);
+
+	g_ptr_array_set_free_func(device->contexts,
+				  (GDestroyNotify)dls_device_delete_context);
 
 	if (device->contexts->len == 0) {
 		if (!under_construction) {
-			DLEYNA_LOG_DEBUG("Last Context lost. Delete device");
-			upnp->lost_server(device->path, upnp->user_data);
-			g_hash_table_remove(upnp->server_udn_map, udn);
+			DLEYNA_LOG_DEBUG("Last Context lost.");
+
+			if (!device->sleeping) {
+				DLEYNA_LOG_DEBUG("Delete device.");
+
+				upnp->lost_server(device->path,
+						  upnp->user_data);
+
+				g_hash_table_remove(upnp->device_udn_map, udn);
+			} else {
+				DLEYNA_LOG_DEBUG("Persist sleeping device.");
+
+				dleyna_task_processor_remove_queues_for_sink(
+						dls_server_get_task_processor(),
+						device->path);
+
+				g_hash_table_insert(
+						upnp->sleeping_device_udn_map,
+						g_strdup(udn),
+						device);
+
+				if (g_hash_table_lookup_extended(
+							  upnp->device_udn_map,
+							  udn,
+							  &key,
+							  &val)) {
+					g_hash_table_steal(upnp->device_udn_map,
+							   udn);
+
+					g_free(key);
+				}
+
+				device->sleeping_context = lost_context;
+
+				lost_context = NULL;
+			}
 		} else {
 			DLEYNA_LOG_WARNING(
 				"Device under construction. Cancelling");
@@ -307,7 +446,7 @@ static void prv_server_unavailable_cb(GUPnPControlPoint *cp,
 			"Device under construction. Switching context");
 
 		/* Cancel previous contruction task chain */
-		g_hash_table_remove(priv_t->upnp->server_uc_map, priv_t->udn);
+		g_hash_table_remove(priv_t->upnp->device_uc_map, priv_t->udn);
 		dleyna_task_queue_set_finally(priv_t->queue_id,
 					      prv_device_context_switch_end);
 		dleyna_task_processor_cancel_queue(priv_t->queue_id);
@@ -330,10 +469,13 @@ static void prv_server_unavailable_cb(GUPnPControlPoint *cp,
 		DLEYNA_LOG_DEBUG("Subscribe on new context");
 
 		device->timeout_id = g_timeout_add_seconds(1,
-				prv_subscribe_to_contents_change,
+				prv_subscribe_to_service_changes,
 				device);
 	}
 
+	if (lost_context != NULL)
+		dls_device_delete_context(lost_context);
+
 on_error:
 
 	DLEYNA_LOG_DEBUG("Exit");
@@ -351,13 +493,13 @@ static void prv_on_context_available(GUPnPContextManager *context_manager,
 
 	cp = gupnp_control_point_new(
 		context,
-		"urn:schemas-upnp-org:device:MediaServer:1");
+		"upnp:rootdevice");
 
 	g_signal_connect(cp, "device-proxy-available",
-			 G_CALLBACK(prv_server_available_cb), upnp);
+			 G_CALLBACK(prv_device_available_cb), upnp);
 
 	g_signal_connect(cp, "device-proxy-unavailable",
-			 G_CALLBACK(prv_server_unavailable_cb), upnp);
+			 G_CALLBACK(prv_device_unavailable_cb), upnp);
 
 	gssdp_resource_browser_set_active(GSSDP_RESOURCE_BROWSER(cp), TRUE);
 	gupnp_context_manager_manage_control_point(upnp->context_manager, cp);
@@ -365,6 +507,7 @@ static void prv_on_context_available(GUPnPContextManager *context_manager,
 }
 
 dls_upnp_t *dls_upnp_new(dleyna_connector_id_t connection,
+			 guint port,
 			 const dleyna_connector_dispatch_cb_t *dispatch_table,
 			 dls_upnp_callback_t found_server,
 			 dls_upnp_callback_t lost_server,
@@ -378,16 +521,21 @@ dls_upnp_t *dls_upnp_new(dleyna_connector_id_t connection,
 	upnp->found_server = found_server;
 	upnp->lost_server = lost_server;
 
-	upnp->server_udn_map = g_hash_table_new_full(g_str_hash, g_str_equal,
-						     g_free,
-						     dls_device_delete);
+	upnp->device_udn_map = g_hash_table_new_full(g_str_hash, g_str_equal,
+						g_free,
+						dls_device_delete);
 
-	upnp->server_uc_map = g_hash_table_new_full(g_str_hash, g_str_equal,
+	upnp->sleeping_device_udn_map = g_hash_table_new_full(g_str_hash,
+						g_str_equal,
+						g_free,
+						dls_device_delete);
+
+	upnp->device_uc_map = g_hash_table_new_full(g_str_hash, g_str_equal,
 						    g_free, NULL);
 
 	dls_prop_maps_new(&upnp->property_map, &upnp->filter_map);
 
-	upnp->context_manager = gupnp_context_manager_create(0);
+	upnp->context_manager = gupnp_context_manager_create(port);
 
 	g_signal_connect(upnp->context_manager, "context-available",
 			 G_CALLBACK(prv_on_context_available),
@@ -399,16 +547,20 @@ dls_upnp_t *dls_upnp_new(dleyna_connector_id_t connection,
 void dls_upnp_delete(dls_upnp_t *upnp)
 {
 	if (upnp) {
+		g_signal_handlers_disconnect_by_func (G_OBJECT (upnp->context_manager),
+						      G_CALLBACK(prv_on_context_available),
+						      upnp);
 		g_object_unref(upnp->context_manager);
 		g_hash_table_unref(upnp->property_map);
 		g_hash_table_unref(upnp->filter_map);
-		g_hash_table_unref(upnp->server_udn_map);
-		g_hash_table_unref(upnp->server_uc_map);
+		g_hash_table_unref(upnp->device_udn_map);
+		g_hash_table_unref(upnp->sleeping_device_udn_map);
+		g_hash_table_unref(upnp->device_uc_map);
 		g_free(upnp);
 	}
 }
 
-GVariant *dls_upnp_get_server_ids(dls_upnp_t *upnp)
+GVariant *dls_upnp_get_device_ids(dls_upnp_t *upnp)
 {
 	GVariantBuilder vb;
 	GHashTableIter iter;
@@ -420,13 +572,20 @@ GVariant *dls_upnp_get_server_ids(dls_upnp_t *upnp)
 
 	g_variant_builder_init(&vb, G_VARIANT_TYPE("ao"));
 
-	g_hash_table_iter_init(&iter, upnp->server_udn_map);
+	g_hash_table_iter_init(&iter, upnp->device_udn_map);
 	while (g_hash_table_iter_next(&iter, NULL, &value)) {
 		device = value;
 		DLEYNA_LOG_DEBUG("Have device %s", device->path);
 		g_variant_builder_add(&vb, "o", device->path);
 	}
 
+	g_hash_table_iter_init(&iter, upnp->sleeping_device_udn_map);
+	while (g_hash_table_iter_next(&iter, NULL, &value)) {
+		device = value;
+		DLEYNA_LOG_DEBUG("Have sleeping device %s", device->path);
+		g_variant_builder_add(&vb, "o", device->path);
+	}
+
 	retval = g_variant_ref_sink(g_variant_builder_end(&vb));
 
 	DLEYNA_LOG_DEBUG("Exit");
@@ -434,9 +593,26 @@ GVariant *dls_upnp_get_server_ids(dls_upnp_t *upnp)
 	return retval;
 }
 
-GHashTable *dls_upnp_get_server_udn_map(dls_upnp_t *upnp)
+GHashTable *dls_upnp_get_device_udn_map(dls_upnp_t *upnp)
 {
-	return upnp->server_udn_map;
+	return upnp->device_udn_map;
+}
+
+GHashTable *dls_upnp_get_sleeping_device_udn_map(dls_upnp_t *upnp)
+{
+	return upnp->sleeping_device_udn_map;
+}
+
+void dls_upnp_delete_sleeping_device(dls_upnp_t *upnp, dls_device_t *device)
+{
+	const char *udn;
+	dls_device_context_t *ctx = device->sleeping_context;
+
+	udn = gupnp_device_info_get_udn((GUPnPDeviceInfo *)ctx->device_proxy);
+
+	upnp->lost_server(device->path, upnp->user_data);
+
+	g_hash_table_remove(upnp->sleeping_device_udn_map, udn);
 }
 
 void dls_upnp_get_children(dls_upnp_t *upnp, dls_client_t *client,
@@ -1088,6 +1264,21 @@ void dls_upnp_get_icon(dls_upnp_t *upnp, dls_client_t *client,
 	DLEYNA_LOG_DEBUG("Exit");
 }
 
+void dls_upnp_wake(dls_upnp_t *upnp, dls_client_t *client,
+		   dls_task_t *task,
+		   dls_upnp_task_complete_t cb)
+{
+	dls_async_task_t *cb_data = (dls_async_task_t *)task;
+
+	DLEYNA_LOG_DEBUG("Enter");
+
+	cb_data->cb = cb;
+
+	dls_device_wake(client, task);
+
+	DLEYNA_LOG_DEBUG("Exit");
+}
+
 void dls_upnp_unsubscribe(dls_upnp_t *upnp)
 {
 	GHashTableIter iter;
@@ -1096,7 +1287,7 @@ void dls_upnp_unsubscribe(dls_upnp_t *upnp)
 
 	DLEYNA_LOG_DEBUG("Enter");
 
-	g_hash_table_iter_init(&iter, upnp->server_udn_map);
+	g_hash_table_iter_init(&iter, upnp->device_udn_map);
 	while (g_hash_table_iter_next(&iter, NULL, &value)) {
 		device = value;
 		dls_device_unsubscribe(device);
@@ -1131,11 +1322,11 @@ gboolean dls_upnp_device_context_exist(dls_device_t *device,
 		goto on_exit;
 
 	/* Check if the device still exist */
-	result = g_hash_table_find(upnp->server_udn_map, prv_device_find,
+	result = g_hash_table_find(upnp->device_udn_map, prv_device_find,
 				   device);
 
 	if (result == NULL)
-		if (g_hash_table_find(upnp->server_uc_map, prv_device_uc_find,
+		if (g_hash_table_find(upnp->device_uc_map, prv_device_uc_find,
 				      device) == NULL)
 			goto on_exit;
 
